# -*- coding: utf-8 -*-
#
# Client for OX' HTTP API
#
# Copyright 2017-2018 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

#
# Client for OX' HTTP API
# API docs:
# * http://oxpedia.org/wiki/index.php?title=HTTP_API
# * http://oxpedia.org/wiki/index.php?title=HTTP_API_MailFilter
#

try:
	from urlparse import urljoin
except ImportError:
	# Python 3
	from urllib.parse import urljoin
import csv
import sys
import time
import json
import logging
from copy import deepcopy

import requests
import univention.admin.uldap as udm_uldap
from univention.config_registry import ConfigRegistry


ucr = ConfigRegistry()
ucr.load()

OX_ATTR_TO_LDAP_ATTR = dict([
	('display_name', 'oxDisplayName'),
	('email1', 'mailPrimaryAddress'),
	('first_name', 'givenName'),
	('login_info', 'uid'),
	('timezone', 'oxTimeZone')
])
LDAP_ATTR_TO_OX_ATTR = dict((v, k) for k, v in OX_ATTR_TO_LDAP_ATTR.items())
# https://documentation.open-xchange.com/components/middleware/http/7.8.4/#column-identifiers ff
OX_API_COLUMN_NAMES = {
	1: 'id',
	2: 'created_by',
	3: 'modified_by',
	20: 'folder_id',
	300: 'title',
	301: 'module',
	302: 'type',
	304: 'subfolders',
	309: 'total',
}
# https://documentation.open-xchange.com/components/middleware/http/7.8.4/#type-of-folder
OX_FOLDER_TYPE = {
	1: 'private',
	2: 'public',
	3: 'shared',
	5: 'system folder',
	7: 'deprecated!',
	16: 'trash',
	20: 'pictures',
	21: 'documents',
	22: 'music',
	23: 'videos',
	24: 'templates',
}

class OxHttpApiError(Exception):
	pass


class NotFoundError(OxHttpApiError):
	pass


class LoginError(OxHttpApiError):
	pass


class AttributeSyncError(OxHttpApiError):
	pass


class ApiError(OxHttpApiError):
	pass


class UnkownMethodError(OxHttpApiError):
	pass


class HandlerError(Exception):
	pass


class OXFolder(object):
	def __init__(self, folder_id, folder_type, title, cache, parent=None):
		self.id = folder_id       # ID of folder
		self.type = folder_type   # https://documentation.open-xchange.com/components/middleware/http/7.8.4/#type-of-folder
		self.title = title        # Name of this folder
		self.parent = parent      # OXFolder, None if root folder
		self.subfolders = list()  # list of OXFolder
		self.resource = dict()    # complete API folder resource
		self.cache = cache        # known folders cache
		self.cache.by_id[self.id] = self

	def __repr__(self):
		return 'OXFolder({folder_id!r}, {folder_type_name!r}, {folder_title!r}, {parent_id!r})'.format(
			folder_id=self.id,
			folder_type_name=OX_FOLDER_TYPE[self.type],
			folder_title=self.title,
			parent_id=self.parent.id if self.parent else None
		)

	@classmethod
	def from_resource(cls, resource, cache):
		obj = cls(
			folder_id=resource['id'],
			folder_type=resource['type'],
			title=resource['title'],
			cache=cache,
			parent=cache.by_id.get(resource['folder_id'])
		)
		obj.resource = resource
		return obj

	def drop_folder_cache(self):
		self.cache = OXFolderCache()

	@property
	def children(self):
		return [f for f in self.cache.by_id.values() if f.parent == self]


class OXFolderCache(object):
	def __init__(self):
		self.by_id = dict()

	@property
	def by_module(self):
		res = dict()
		for folder_id, folder in self.by_id.items():
			res.setdefault(folder.resource.get('module'), []).append(folder)
		return res


def lists_to_dicts(response, columns):
	"""
	Convert a list of lists (OX API response) to a list of dicts.

	:param response: list of lists (response of API calls)
	:param columns: str (column ids as in API calls)
	:return: list of dicts
	"""
	names = [OX_API_COLUMN_NAMES[int(c)] for c in columns.split(',')]
	return [dict(zip(names, r)) for r in response]


class Client(object):
	LOG_REQUEST = 5
	LOG_RESPONSE = 4
	RAW_RESPONSE_HANLDER_CHUNK_SIZE = 2048

	def __init__(self, name, password, server=None, ssl=True, *args, **kwargs):
		"""
		Log user in.

		:param name: str: username for authentication
		:param password: str: password for authentication
		:param server: str: FQDN of server to connect to, FQDN of local server if None
		:param ssl: bool: whether to connect using HTTP or HTTPS
		:param kwargs:
		* log_level: logging.{INFO,DEBUG} or LOG_REQUEST>LOG_RESPONSE to log API requests & responses (default: logging.INFO)
		* log_target: stream or file name (default: sys.stderr)
		* log_kwargs: arguments to pass to logging handler
		"""
		assert isinstance(name, basestring)
		assert isinstance(password, basestring)
		self.username = name
		self.password = password
		self.server = server or '{}.{}'.format(ucr['hostname'], ucr['domainname'])
		self.session = None
		self.cookies = None
		self._lo = None
		self.base_url = 'http{}://{}/ajax/'.format('s' if ssl else '', self.server)
		self.log_level = kwargs.get('log_level', logging.INFO)
		self.log_target = kwargs.get('log_target', sys.stderr)
		self.logger = self._get_logger(self.log_level, self.log_target, **kwargs.get('log_kwargs', {}))
		self.folder_cache = OXFolderCache()

		try:
			self.login()
		except LoginError as exc:
			if 'Please try again later' in str(exc):
				self.logger.error('Error logging in. Sleeping 10s and retrying once. Error was:\n%s', str(exc))
				time.sleep(10)
				self.login()
			else:
				raise

	@classmethod
	def _get_logger(cls, log_level, log_target, **kwargs):
		if not hasattr(logging, 'LOG_REQUEST'):
			logging.addLevelName(cls.LOG_REQUEST, 'REQUEST')
		if not hasattr(logging, 'LOG_RESPONSE'):
			logging.addLevelName(cls.LOG_RESPONSE, 'RESPONSE')
		if log_level > cls.LOG_REQUEST:
			logger_name = 'ox_http_api_client'
		else:
			# capture requests lib output
			logger_name = None
		logger = logging.getLogger(logger_name)
		logger.setLevel(log_level)
		logger.request = lambda msg, *args, **kwargs: logger.log(cls.LOG_REQUEST, msg, *args, **kwargs)
		logger.response = lambda msg, *args, **kwargs: logger.log(cls.LOG_RESPONSE, msg, *args, **kwargs)
		my_handler_name = 'ox_http_api_client'
		if not any(handler.get_name() == my_handler_name for handler in logger.handlers):
			if isinstance(log_target, file) or hasattr(log_target, "write"):
				handler = logging.StreamHandler(log_target)
			else:
				handler = logging.FileHandler(log_target, **kwargs)
			handler.set_name(my_handler_name)
			if log_level < logging.INFO:
				fmt = '%(asctime)s %(levelname)-5s %(module)s.%(funcName)s:%(lineno)d  %(message)s'
			else:
				fmt = '%(asctime)s %(levelname)s: %(message)s'
			handler.setFormatter(logging.Formatter(
				fmt=fmt,
				datefmt='%Y-%m-%d %H:%M:%S'))
			handler.setLevel(log_level)
			logger.addHandler(handler)
		return logger

	@staticmethod
	def _get_json(obj):
		"""
		Handle different requests versions.

		:param obj: response object
		:return: dict: dict from parsed json
		"""
		if callable(obj.json):
			return obj.json()
		else:
			return obj.json

	def login(self):
		url = urljoin(self.base_url, 'login?action=login')
		data = dict(name=self.username, password=self.password)
		_data = data.copy()
		_data['password'] = 'x' * len(_data['password'])
		self.logger.request('Logging in: POST(url={!r} data={!r})'.format(url, _data))
		response = requests.post(url, data=data)
		if response.status_code != 200:
			raise LoginError('Received status_code={} when trying to log in: {}'.format(
				response.status_code,
				response.content))
		self.logger.response(response.content)
		json_response = self._get_json(response)
		if 'error' in json_response:
			raise LoginError('Error logging in: {}'.format(self._get_json(response)['error']))

		self.logger.info('Logged in as user {!r}.'.format(self.username))
		self.session = json_response['session']
		self.cookies = response.cookies

	def logout(self):
		self.get('login?action=logout', response_handler=self._raw_response_handler)
		self.session = self.cookies = None
		self.logger.info('Logged out user {!r}.'.format(self.username))

	def _call_api(self, method, url_end, data=None, data_handler=None, response_handler=None, **params):
		url = urljoin(self.base_url, url_end)
		params.update(dict(session=self.session))
		kwargs = dict(url=url, params=params, cookies=self.cookies)
		if not data_handler:
			data_handler = self._json_data_handler
		if not response_handler:
			response_handler = self._json_response_handler
		if data:
			kwargs.update(data_handler(data))
		self.logger.request('%r(%r)', method, kwargs)
		meth = getattr(requests, method)
		response = meth(**kwargs)
		self.logger.response(response.content)
		if response.status_code != 200:
			raise ApiError('Received status_code={!r} with reason={!r} for {!r}({!r}).'.format(response.status_code, getattr(response, 'reason'), method, kwargs))
		try:
			return response_handler(response)
		except HandlerError as exc:
			raise ApiError('When calling {} {!r} with params={!r} and data={}: {}'.format(
				method.upper(), url, params, data, exc))

	@classmethod
	def _file_data_handler(cls, file_handle):
		"""
		:param file_handle: file like object to read data from
		:return: dict: use as data argument for api calls
		"""
		# OX wants the filename to be 'file', not basename(data)
		return {'files': {'file': ('file', file_handle.read(), 'text/plain')}}

	@classmethod
	def _json_data_handler(cls, data):
		return {'data': json.dumps(data)}

	@classmethod
	def _json_response_handler(cls, response):
		"""
		:param response: response object of request
		:return: str: JSON
		"""
		json_response = cls._get_json(response)
		try:
			return json_response['data']
		except KeyError:
			if 'error_desc' in json_response or 'error' in json_response:
				raise HandlerError(json_response.get('error_desc', json_response['error']))
			else:
				return json_response

	@classmethod
	def _csv_response_handler(cls, response):
		"""
		:param response: response object of request
		:return: tuple(str, generator): encoding, csv.DictReader instance
		"""
		return response.encoding, csv.DictReader(response.iter_lines())

	@classmethod
	def _raw_response_handler(cls, response):
		"""
		Set Client.RAW_RESPONSE_HANLDER_CHUNK_SIZE to an int if you want to
		change the chunks size (default: 2048).

		:param response: response object of request
		:return: tuple(str, generator): encoding, iterator over the content
		"""
		return response.encoding, response.iter_content(chunk_size=cls.RAW_RESPONSE_HANLDER_CHUNK_SIZE)

	def get_json(self, url_end, **params):
		"""
		Get JSON response from OX API.

		:param url_end: str: API endpoint, e.g. ''
		:param params: dict: parameters to pass to OX API
		:return: dict: response data
		"""
		return self._call_api('get', url_end, **params)

	get = get_json

	def get_raw(self, url_end, **params):
		"""
		Get raw text from export module.

		:param url_end: str: API endpoint
		:param params: dict: parameters to pass to OX API
		:return: tuple(str, generator(str)): encoding, text
		"""
		return self._call_api('get', url_end, response_handler=self._raw_response_handler, **params)

	def get_csv(self, url_end, **params):
		"""
		Get CSV from export module.

		:param url_end: str: API endpoint, e.g. 'export?action=CSV'
		:param params: dict: parameters to pass to OX API
		:return: tuple(str: encoding, csv.DictReader instance)
		"""
		return self._call_api('get', url_end, response_handler=self._csv_response_handler, **params)

	get_ical = get_raw
	get_vcard = get_raw

	def post(self, url_end, data, **params):
		return self._call_api('post', url_end, data, **params)

	def post_ical(self, url_end, data, **params):
		assert 'folder' in params, 'Missing parameter "folder" in params.'

		return self._call_api('post', url_end, data, data_handler=self._file_data_handler, response_handler=self._raw_response_handler, **params)

	def post_csv(self, url_end, data, **params):
		assert 'folder' in params, 'Missing parameter "folder" in params.'
		assert 'charset' in params, 'Missing parameter "charset" in params.'

		return self._call_api('post', url_end, data, data_handler=self._file_data_handler, response_handler=self._raw_response_handler, **params)

	def post_vcard(self, url_end, data, **params):
		assert 'folder' in params, 'Missing parameter "folder" in params.'
		# TODO
		pass

	def put(self, url_end, data=None, **params):
		return self._call_api('put', url_end, data, **params)

	def get_context_id(self):
		"""
		Get the context ID of the logged in user.

		:return: dict: {u'data': 10}
		"""
		return self.get('config/context_id')

	def get_user(self, oid=None):
		"""

		:param oid: int: Object ID of the requested user, current user if None
		:return: dict
		"""
		kwargs = dict(action='get')
		if oid:
			kwargs['id'] = oid
		return self.user(**kwargs)

	def get_folder(self, folder_id, recursive=False):
		"""
		Fetch data about a folder.

		:param folder_id: str
		:param recursive: bool: if it has subfolders, fetch those too
		:return: OXFolder object
		"""
		self.logger.debug('get_folder(%r, %r)', folder_id, recursive)
		folder_resource = self.folders('get', id=folder_id)
		ox_folder = OXFolder.from_resource(folder_resource, self.folder_cache)
		if recursive and ox_folder.resource['subfolders']:
			ox_folder.subfolders = [self.get_folder(sf_id, True) for sf_id in self.get_subfolder_ids(folder_id)]
		return ox_folder

	def get_subfolder_ids(self, folder_id):
		"""
		Get list of IDs of subfolders of a folder.

		:param folder_id:
		:return: list of str
		"""
		subfolders = self.folders('list', parent=folder_id, columns='1', all=1)
		return [sf[0] for sf in subfolders]

	def get_all_folders(self):
		"""
		Fetch data about all of the users folders and represent in a tree
		structure of OXFolder objects, starting with the root folders.

		:return: lists of OXFolder objects
		"""
		root_folder_ids = [sf[0] for sf in self.folders('root', columns='1')]
		return [self.get_folder(rf_id, recursive=True) for rf_id in root_folder_ids]

	@classmethod
	def format_folder_structure(cls, folder, indent=0):
		"""
		Tree-like view of folder objects.

		:param folder: OXFolder: start with this folder
		:param indent: int: spaces to prepend
		:return: str
		"""
		res = '{indent}{folder}'.format(indent=' ' * indent, folder=folder)
		for sf in folder.subfolders:
			res += '\n{}'.format(cls.format_folder_structure(sf, indent + 4))
		return res

	def list_users(self, columns=None):
		"""
		To find IDs of columns look at:
		- http://oxpedia.org/wiki/index.php?title=HTTP_API#CommonObjectData
		- http://oxpedia.org/wiki/index.php?title=HTTP_API#DetailedContactData
		- http://oxpedia.org/wiki/index.php?title=HTTP_API#DetailedUserData

		:param columns: list: list of ints (IDs of columns to show)
		:return: list of lists
		"""
		if columns:
			assert isinstance(columns, list)
		else:
			columns = [1, 4, 223, 500, 501, 502, 524]
		return self.user('all', columns=",".join(str(x) for x in columns))

	def list_groups(self, pattern='*'):
		"""
		Get all groups matching pattern.

		:param pattern: str: search pattern
		:return: list of lists
		"""
		return self.group('search', data=dict(pattern=pattern))

	def export(self, action, folder_id, columns=None, export_dlists=None, raw=False, **kwargs):
		"""
		The OX HTTP API 'export' function for contact, task and calendar data.

		:param action: str: one of 'CSV', 'ICAL', 'VCARD', 'RAW'
		:param folder_id: int/str: ObjectID of the calendar/contact folder whose contents
		are to be exported.
		:param columns: list: optional, only for CSV
		:param export_dlists: bool: optional, only for CSV
		:param raw: bool: if True: force return of raw text
		:return: tuple(str: encoding, either for ICAL|VCARD a str generator or for CSV a csv.DictReader instance)
		"""
		if action not in ('CSV', 'ICAL', 'VCARD'):
			raise UnkownMethodError('Unknown action {!r} for export.'.format(action))

		if raw:
			method = self.get_raw
		else:
			method = {
				'CSV': self.get_csv,
				'ICAL': self.get_ical,
				'VCARD': self.get_vcard,
			}[action]

		kwargs = deepcopy(kwargs)
		kwargs.update(dict(
			url_end='export?action={}'.format(action),
			folder=folder_id,
		))

		if columns:
			kwargs['columns'] = columns
		if export_dlists:
			kwargs['export_dlists'] = export_dlists

		self.logger.debug('Exporting folder with id %r...', folder_id)
		return method(**kwargs)

	def export_calendar(self, folder_id):
		"""
		Export calendar data.
		https://documentation.open-xchange.com/components/middleware/http/7.8.4/#!/Export/exportAsICal

		:param folder_id: int/str: ObjectID of the calendar folder into which to import
		:return: tuple(str: encoding, generator with the iCal raw text (MIME type text/calendar))
		"""
		return self.export('ICAL', folder_id=folder_id)

	def export_contacts(self, folder_id, raw=False):
		"""
		Export contact data and optionally parse it.
		https://documentation.open-xchange.com/components/middleware/http/7.8.4/#!/Export/exportAsCSV

		:param folder_id: int/str: ObjectID of the calendar folder into which to import
		:param raw: bool: if True: return raw text, if False: return a csv.DictReader instance
		:return: tuple(str: encoding, generator with the raw text|csv.DictReader instance)
		"""
		return self.export('CSV', folder_id=folder_id, raw=raw)

	def folders(self, action, data=None, **params):
		"""
		Access/change OX drive folder information.

		List all folders with IDs: folders('root', columns='1,300')

		:param action: str: GET: 'root', 'list', 'path', 'updates', 'get', 'allVisible', 'shares'
		PUT: 'update', 'new', 'delete', 'clear', 'notify'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('root', 'list', 'path', 'updates', 'get', 'allVisible', 'shares'):
			method = self.get
		elif action in ('update', 'new', 'delete', 'clear', 'notify'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action {!r} for contacts.'.format(action))
		return method('folders?action={}'.format(action), data=data, **params)

	def folder_new(self, folder_id, title, folder_module, folder_data=None, **kwargs):
		"""
		Create a new folder.

		:param folder_id: str: object ID of the parents folder
		:param title: str: name
		:param folder_module: str: 'tasks', calendar', 'contacts', 'infostore' or 'mail'
		:param folder_data: dict: further folder data (see
		https://documentation.open-xchange.com/components/middleware/http/7.8.4/#!/Folders/createFolder)
		:return: dict: a dict incl. the ID of the new folder
		"""
		if not folder_data:
			folder_data = dict()
		folder_data.update({'folder': {'title': title, 'module': folder_module}})
		return self.folders('new', folder_id=folder_id, data=folder_data, **kwargs)

	def contacts(self, action, data=None, **params):
		"""
		Access/change contact information.

		Ex: list all global contacts: contacts_get('all', folder=6, columns='1,20,223,500,501,502')

		:param action: str: GET: 'all', 'updates', 'get', 'getuser', 'anniversaries', 'birthdays', 'autocomplete'
		PUT: 'list', 'listuser', 'update', 'new', 'delete', 'search', 'advancedSearch'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('all', 'updates', 'get', 'getuser', 'anniversaries', 'birthdays', 'autocomplete'):
			method = self.get
		elif action in ('list', 'listuser', 'update', 'new', 'delete', 'search', 'advancedSearch'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action {!r} for contacts.'.format(action))
		return method('contacts?action={}'.format(action), data=data, **params)

	def user(self, action, data=None, **params):
		"""
		Access/change user information.

		:param action: str: GET: 'all', 'get', 'getAttribute'
		PUT: 'list', 'update', 'search', 'setAttribute'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('all', 'get', 'getAttribute'):
			method = self.get
		elif action in ('list', 'update', 'search', 'setAttribute'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action {!r} for user.'.format(action))
		return method('user?action={}'.format(action), data=data, **params)

	def group(self, action, data=None, **params):
		"""
		Access/change group information.
		http://oxpedia.org/wiki/index.php?title=HTTP_API#Module_.22groups.22

		:param action: str: GET: 'get', 'updates'
		PUT: 'list', 'update', 'search', 'new', 'delete'
		:param data: dict/bytes/file: dict, bytes, or file-like object to send in the body of a PUT request
		:param params: parameters to pass to OX API
		:return: list of lists
		"""
		if action in ('get', 'updates'):
			method = self.get
		elif action in ('list', 'update', 'search', 'new', 'delete'):
			method = self.put
		else:
			raise UnkownMethodError('Unknown action {!r} for group.'.format(action))
		return method('group?action={}'.format(action), data=data, **params)

	def import_data(self, action, folder_id, import_file, **kwargs):
		"""
		The OX HTTP API 'import' function for contact, task and calendar data.

		:param action: str: one of 'CSV', 'ICAL', 'VCARD', 'OUTLOOK_CSV'
		:param folder_id: int/str: ObjectID of the calendar/contact folder whose contents
		are to be exported.
		:param import_file: file like object to read import data from
		:param kwargs: dict: arguments to pass to api call
		:return: None
		"""
		if action not in ('CSV', 'ICAL', 'VCARD', 'OUTLOOK_CSV'):
			raise UnkownMethodError('Unknown action {!r} for import_data.'.format(action))

		method = {
			'CSV': self.post_csv,
			'ICAL': self.post_ical,
			'VCARD': self.post_vcard,
			'OUTLOOK_CSV': self.post_csv,
		}[action]

		kwargs = deepcopy(kwargs)
		kwargs.update(dict(
			url_end='import?action={}'.format(action),
			folder=folder_id,
			data=import_file,
		))
		self.logger.debug('kwargs=%r', kwargs)
		return method(**kwargs)

	def import_calendar(self, folder_id, import_file, suppress_notification=True, ignore_uids=True, **kwargs):
		"""
		Import calendar data.
		https://documentation.open-xchange.com/components/middleware/http/7.8.4/#!/Import/importICal

		:param folder_id: int/str: ObjectID of the calendar folder into which to import
		:param import_file: file like object to read import data from
		:param suppress_notification: bool: disable the notifications for new appointments
		:param ignore_uids: bool: partially ignore UIDs and replace with random ones
		:param kwargs: dict: arguments to pass to api call
		:return: None
		"""
		kwargs = deepcopy(kwargs)
		kwargs.update(dict(
			action='ICAL',
			folder_id=folder_id,
			import_file=import_file,
			suppressNotification=suppress_notification,
			ignoreUIDs=ignore_uids,
		))
		return self.import_data(**kwargs)

	def import_contacts(self, folder_id, import_file, charset, **kwargs):
		"""
		The OX HTTP API 'import' function for contact, task and calendar data.
		https://documentation.open-xchange.com/components/middleware/http/7.8.4/#!/Import/importCSV

		:param action: str: one of 'CSV', 'ICAL', 'VCARD', 'OUTLOOK_CSV'
		:param folder_id: int/str: ObjectID of the contact folder into which to import
		:param import_file: file like object to read import data from
		:param charset: str: character set in which the data in import_file is encoded
		:param kwargs: dict: arguments to pass to api call
		:return: None
		"""
		kwargs = deepcopy(kwargs)
		kwargs.update(dict(
			action='CSV',
			folder_id=folder_id,
			import_file=import_file,
			charset=charset,
		))
		return self.import_data(**kwargs)

	def username_to_id(self, username):
		"""
		Get the user ID (not object ID) of a user.

		:param username: str: surname
		:return: int: ID or NotFoundError
		"""
		assert isinstance(username, basestring)
		user_list = self.post('user?action=all', data=username, columns="1,615")
		for oid, _username in user_list:
			if _username == username:
				return oid
		raise NotFoundError('Could not get object ID of user {!r}.'.format(username))

	@property
	def lo(self):
		if not self._lo:
			self._lo = udm_uldap.getMachineConnection(ldap_master=True)[0]
		return self._lo

	def compare_ldap_and_ox_user(self, username):
		oid = self.username_to_id(username)
		self.logger.info('User {!r} has object ID {!r}.'.format(username, oid))

		ox_user = self.get_user(oid)
		_dn, ldap_user = self.lo.search(filter='(&(objectClass=posixAccount)(uid={}))'.format(username))[0]
		for ox_attr, ldap_attr in OX_ATTR_TO_LDAP_ATTR.items():
			if ox_user[ox_attr] != ldap_user[ldap_attr][0]:
				raise AttributeSyncError('Attributes not equal: ox_user[{}]={} ldap_user[][0]={}'.format(
					ox_attr, ox_user[ox_attr], ldap_attr, ldap_user[ldap_attr][0]))
		self.logger.info('OK: Attributes in LDAP and in OX-DB synced correctly.')
